#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import asyncio
import dataclasses
import json
import logging
import time
from asyncio.exceptions import TimeoutError

import backoff
import websockets
from websockets.exceptions import (
    ConnectionClosedError,
    ConnectionClosedOK,
    InvalidStatusCode,
)


from connect.eaas.config import ConfigHelper
from connect.eaas.constants import (
    DELAY_ON_CONNECT_EXCEPTION_SECONDS,
    MAX_RETRY_DELAY_TIME_SECONDS,
    MAX_RETRY_TIME_GENERIC_SECONDS,
    MAX_RETRY_TIME_MAINTENANCE_SECONDS,
    RESULT_SENDER_MAX_RETRIES,
    RESULT_SENDER_WAIT_GRACE_SECONDS,
)
from connect.eaas.dataclasses import (
    CapabilitiesPayload,
    Message,
    MessageType,
    parse_message,
)
from connect.eaas.exceptions import (
    CommunicationError,
    MaintenanceError,
    StopBackoffError,
)
from connect.eaas.handler import ExtensionHandler
from connect.eaas.helpers import get_version, to_ordinal
from connect.eaas.managers import (
    BackgroundTasksManager,
    InteractiveTasksManager,
    ScheduledTasksManager,
)


logger = logging.getLogger(__name__)


def _get_max_retry_time_maintenance():
    return MAX_RETRY_TIME_MAINTENANCE_SECONDS


def _get_max_retry_time_generic():
    return MAX_RETRY_TIME_GENERIC_SECONDS


def _get_max_retry_delay_time():
    return MAX_RETRY_DELAY_TIME_SECONDS


class Worker:
    """
    The Worker is responsible to handle the websocket connection
    with the server. It will send the extension capabilities to
    the server and wait for tasks that need to be processed using
    the tasks manager.
    """
    def __init__(self, secure=True, runner_type=None):
        self.runner_type = runner_type
        self.config = ConfigHelper(secure)
        self.handler = ExtensionHandler(self.config)
        self.lock = asyncio.Lock()
        self.results_queue = asyncio.Queue()
        self.run_event = asyncio.Event()
        self.stop_event = asyncio.Event()
        self.background_manager = BackgroundTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )
        self.interactive_manager = InteractiveTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )
        self.scheduled_manager = ScheduledTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )
        self.ws = None
        self.main_task = None
        self.results_task = None

    @property
    def running_tasks(self):
        return (
            self.background_manager.running_tasks
            + self.interactive_manager.running_tasks
            + self.scheduled_manager.running_tasks
        )

    def get_url(self):
        url = self.config.get_ws_url()
        url = f'{url}?running_tasks={self.background_manager.running_tasks}'
        url = f'{url}&running_scheduled_tasks={self.scheduled_manager.running_tasks}'
        if self.runner_type:
            url = f'{url}&runner_type={self.runner_type}'
        return url

    async def ensure_connection(self):  # noqa: CCR001
        """
        Ensure that a websocket connection is established.
        """
        @backoff.on_exception(
            backoff.expo,
            CommunicationError,
            max_time=_get_max_retry_time_generic,
            max_value=_get_max_retry_delay_time,
            on_backoff=self._backoff_log,
            giveup=self._backoff_shutdown,
        )
        @backoff.on_exception(
            backoff.expo,
            MaintenanceError,
            max_time=_get_max_retry_time_maintenance,
            max_value=_get_max_retry_delay_time,
            on_backoff=self._backoff_log,
            giveup=self._backoff_shutdown,
        )
        async def _connect():
            async with self.lock:
                if self.ws is None or not self.ws.open:
                    try:
                        url = self.get_url()
                        self.ws = await websockets.connect(
                            url,
                            extra_headers=self.config.get_headers(),
                            ping_interval=60,
                            ping_timeout=60,
                            max_queue=128,
                        )
                        await (await self.ws.ping())
                        await self.do_handshake()
                        logger.info(f'Connected to {url}')
                    except InvalidStatusCode as ic:
                        if ic.status_code == 502:
                            logger.warning('Maintenance in progress...')
                            raise MaintenanceError()
                        else:
                            logger.warning(
                                f'Received an unexpected status from server: {ic.status_code}...',
                            )
                            raise CommunicationError()
                    except Exception as e:
                        logger.warning(f'Received an unexpected exception: {e}...')
                        raise CommunicationError()

        await _connect()

    async def do_handshake(self):
        await self.send(self.get_capabilities())
        message = await asyncio.wait_for(self.ws.recv(), timeout=5)
        message = parse_message(json.loads(message))
        await self.process_configuration(message.data)

    async def send(self, message):
        """
        Send a message to the websocket server.
        """
        await self.ws.send(json.dumps(message))

    async def receive(self):
        """
        Receive a message from the websocket server.
        """
        try:
            message = await asyncio.wait_for(self.ws.recv(), timeout=1)
            # message = await self.ws.recv()
            return json.loads(message)
        except TimeoutError:  # pragma: no cover
            pass

    def get_capabilities(self):
        return dataclasses.asdict(
            Message(
                message_type=MessageType.CAPABILITIES,
                data=CapabilitiesPayload(
                    self.handler.capabilities,
                    self.handler.variables,
                    self.handler.schedulables,
                    self.handler.readme,
                    self.handler.changelog,
                    get_version(),
                ),
            ),
        )

    async def run(self):  # noqa: CCR001
        """
        Main loop for the websocket connection.
        Once started, this worker will send the capabilities message to
        the websocket server and start a loop to receive messages from the
        websocket server.
        """
        await self.run_event.wait()
        while self.run_event.is_set():
            try:
                await self.ensure_connection()
                while self.run_event.is_set():
                    message = await self.receive()
                    if not message:
                        continue
                    logger.debug('New message received via WS')
                    await self.process_message(message)
            except (ConnectionClosedOK, StopBackoffError):
                self.run_event.clear()
                continue
            except (CommunicationError, MaintenanceError):
                logger.error('Max connection attemps reached, exit!')
                self.run_event.clear()
                continue
            except ConnectionClosedError:
                logger.warning(
                    f'Disconnected from: {self.get_url()}'
                    f', try to reconnect in {DELAY_ON_CONNECT_EXCEPTION_SECONDS}s',
                )
                await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
            except InvalidStatusCode as ic:
                if ic.status_code == 502:
                    logger.warning(
                        'Maintenance in progress'
                        f', try to reconnect in {DELAY_ON_CONNECT_EXCEPTION_SECONDS}s',
                    )
                    await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
                else:
                    logger.warning(
                        f'Received an unexpected status from server: {ic.status_code}'
                        f', try to reconnect in {DELAY_ON_CONNECT_EXCEPTION_SECONDS}s',
                    )
                    await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
            except Exception as e:
                logger.exception(
                    f'Unexpected exception {e}'
                    f', try to reconnect in {DELAY_ON_CONNECT_EXCEPTION_SECONDS}s',
                )
                await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
        logger.info('Consumer loop exited!')

    async def process_message(self, data):
        """
        Process a message received from the websocket server.
        """
        message = parse_message(data)
        if message.message_type == MessageType.CONFIGURATION:
            await self.process_configuration(message.data)
        elif message.message_type == MessageType.TASK:
            await self.process_task(message.data)
        elif message.message_type == MessageType.SHUTDOWN:
            await self.shutdown()

    async def process_task(self, task_data):
        """Send a task to a manager based on task category."""
        logger.info(f'received new {task_data.task_category} task: {task_data.task_id}')
        manager = getattr(self, f'{task_data.task_category}_manager')
        await manager.submit(task_data)
        logger.info(f'task {task_data.task_id} submitted for processing')

    async def result_sender(self):  # noqa: CCR001
        """
        Dequeues results from the results queue and send it to
        the EaaS backend.
        """
        await self.run_event.wait()
        while True:
            if self.results_queue.empty():
                if not self.run_event.is_set() and self.running_tasks == 0:
                    logger.info('Worker exiting and no more running tasks: exit!')
                    return
                await asyncio.sleep(.5)
                continue
            logger.info(
                f'Current processing status: running={self.running_tasks} '
                f'results={self.results_queue.qsize()}',
            )
            result = await self.results_queue.get()
            logger.info(f'Got a result from queue: {result.task_id}')
            retries = 0
            while retries < RESULT_SENDER_MAX_RETRIES:
                try:
                    message = Message(
                        message_type=MessageType.TASK,
                        data=result,
                    )
                    await self.ensure_connection()
                    await self.send(dataclasses.asdict(message))
                    logger.info(f'Result for task {result.task_id} has been sent.')
                    break
                except Exception:
                    logger.warning(
                        f'Attemp {retries} to send results for task {result.task_id} has failed.',
                    )
                    retries += 1
                    await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
            else:
                logger.warning(
                    f'Max retries exceeded ({RESULT_SENDER_MAX_RETRIES})'
                    f' for sending results of task {result.task_id}',
                )

            if not self.run_event.is_set():
                logger.info(
                    f'Current processing status: running={self.running_tasks} '
                    f'results={self.results_queue.qsize()}',
                )

    async def process_configuration(self, data):
        """
        Process the configuration message.
        It will stop the tasks manager so the extension can be
        reconfigured, then restart the tasks manager.
        """
        self.config.update_dynamic_config(data)
        logger.info('Extension configuration has been updated.')

    async def shutdown(self):
        """
        Shutdown the extension runner.
        """
        logger.info('Shutdown extension runner.')
        self.stop()

    async def start(self):
        """
        Start the runner.
        """
        logger.info('Starting control worker...')
        self.main_task = asyncio.create_task(self.run())
        self.results_task = asyncio.create_task(self.result_sender())
        self.run_event.set()
        logger.info('Control worker started')
        await self.stop_event.wait()
        result_timeout = self.config.get_timeout('background') + RESULT_SENDER_WAIT_GRACE_SECONDS
        try:
            await asyncio.wait_for(
                self.results_task,
                timeout=result_timeout,
            )
        except asyncio.TimeoutError:
            logger.error(
                f'Cannot send all results timeout of {result_timeout} exceeded, cancel task',
            )
            self.results_task.cancel()
            try:
                await self.results_task
            except asyncio.CancelledError:
                logger.info('Result sender task has been cancelled')

        await self.main_task
        if self.ws:
            await self.ws.close()
        logger.info('Control worker stopped')

    def stop(self):
        """
        Stop the runner.
        """
        logger.info('Stopping control worker...')
        self.run_event.clear()
        self.stop_event.set()

    async def send_shutdown(self):
        msg = Message(MessageType.SHUTDOWN)
        await self.send(dataclasses.asdict(msg))

    def handle_signal(self):
        asyncio.create_task(self.send_shutdown())
        time.sleep(1)
        self.stop()

    def _backoff_shutdown(self, _):
        if not self.run_event.is_set():
            logger.info('Worker exiting, stop backoff loop')
            raise StopBackoffError()

    def _backoff_log(self, details):
        logger.info(
            f'{to_ordinal(details["tries"])} communication attempt failed, backing off waiting '
            f'{details["wait"]:.2f} seconds after next retry. '
            f'Elapsed time: {details["elapsed"]:.2f}'
            ' seconds.',
        )



import dataclasses
from typing import Any, Optional, Union


class TaskType:
    ASSET_PURCHASE_REQUEST_PROCESSING = 'asset_purchase_request_processing'
    ASSET_CHANGE_REQUEST_PROCESSING = 'asset_change_request_processing'
    ASSET_SUSPEND_REQUEST_PROCESSING = 'asset_suspend_request_processing'
    ASSET_RESUME_REQUEST_PROCESSING = 'asset_resume_request_processing'
    ASSET_CANCEL_REQUEST_PROCESSING = 'asset_cancel_request_processing'
    ASSET_ADJUSTMENT_REQUEST_PROCESSING = 'asset_adjustment_request_processing'
    ASSET_PURCHASE_REQUEST_VALIDATION = 'asset_purchase_request_validation'
    ASSET_CHANGE_REQUEST_VALIDATION = 'asset_change_request_validation'
    PRODUCT_ACTION_EXECUTION = 'product_action_execution'
    PRODUCT_CUSTOM_EVENT_PROCESSING = 'product_custom_event_processing'
    TIER_CONFIG_SETUP_REQUEST_PROCESSING = 'tier_config_setup_request_processing'
    TIER_CONFIG_CHANGE_REQUEST_PROCESSING = 'tier_config_change_request_processing'
    TIER_CONFIG_ADJUSTMENT_REQUEST_PROCESSING = 'tier_config_adjustment_request_processing'
    TIER_CONFIG_SETUP_REQUEST_VALIDATION = 'tier_config_setup_request_validation'
    TIER_CONFIG_CHANGE_REQUEST_VALIDATION = 'tier_config_change_request_validation'
    SCHEDULED_EXECUTION = 'scheduled_execution'
    LISTING_NEW_REQUEST_PROCESSING = 'listing_new_request_processing'
    LISTING_REMOVE_REQUEST_PROCESSING = 'listing_remove_request_processing'
    TIER_ACCOUNT_UPDATE_REQUEST_PROCESSING = 'tier_account_update_request_processing'
    USAGE_FILE_REQUEST_PROCESSING = 'usage_file_request_processing'
    PART_USAGE_FILE_REQUEST_PROCESSING = 'part_usage_file_request_processing'


class MessageType:
    CAPABILITIES = 'capabilities'
    CONFIGURATION = 'configuration'
    TASK = 'task'
    SHUTDOWN = 'shutdown'


class TaskCategory:
    BACKGROUND = 'background'
    INTERACTIVE = 'interactive'
    SCHEDULED = 'scheduled'


class ResultType:
    SUCCESS = 'success'
    RESCHEDULE = 'reschedule'
    SKIP = 'skip'
    RETRY = 'retry'
    FAIL = 'fail'


@dataclasses.dataclass
class TaskPayload:
    task_id: str
    task_category: str
    task_type: str
    object_id: str
    result: Optional[str] = None
    data: Optional[Any] = None
    countdown: int = 0
    output: Optional[str] = None
    correlation_id: Optional[str] = None
    reply_to: Optional[str] = None
    runtime: Optional[float] = 0.0


@dataclasses.dataclass
class ConfigurationPayload:
    configuration: Optional[dict] = None
    logging_api_key: Optional[str] = None
    environment_type: Optional[str] = None
    account_id: Optional[str] = None
    account_name: Optional[str] = None
    service_id: Optional[str] = None
    log_level: Optional[str] = None
    runner_log_level: Optional[str] = None
    product_id: Optional[str] = None
    hub_id: Optional[str] = None


@dataclasses.dataclass
class CapabilitiesPayload:
    capabilities: dict
    variables: Optional[list] = None
    schedulables: Optional[list] = None
    readme_url: Optional[str] = None
    changelog_url: Optional[str] = None
    runner_version: Optional[str] = None


@dataclasses.dataclass
class Message:
    message_type: str
    data: Optional[Union[CapabilitiesPayload, ConfigurationPayload, TaskPayload]] = None


def from_dict(cls, data):
    field_names = {f.name for f in dataclasses.fields(cls)}
    return cls(**{k: v for k, v in data.items() if k in field_names})


def parse_message(payload):
    message_type = payload['message_type']
    if message_type == MessageType.CONFIGURATION:
        data = from_dict(ConfigurationPayload, payload.get('data'))
    elif message_type == MessageType.TASK:
        data = from_dict(TaskPayload, payload.get('data'))
    elif message_type == MessageType.CAPABILITIES:
        data = from_dict(CapabilitiesPayload, payload.get('data'))
    else:
        data = payload.get('data')

    return Message(message_type=message_type, data=data)



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import argparse
import asyncio
import logging
import logging.config
import signal
import threading
import time
from multiprocessing import Process

import uvloop

from connect.eaas.helpers import notify_process_restarted
from connect.eaas.worker import Worker


logger = logging.getLogger('connect.eaas')

PROCESS_CHECK_INTERVAL_SECS = 5


def configure_logger(debug):
    logging.config.dictConfig(
        {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '%(asctime)s %(name)s %(levelname)s PID_%(process)d %(message)s',
                },
            },
            'filters': {},
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose',
                },
                'null': {
                    'class': 'logging.NullHandler',
                },
            },
            'loggers': {
                'connect.eaas': {
                    'handlers': ['console'],
                    'level': 'DEBUG' if debug else 'INFO',
                },
                'eaas.extension': {
                    'handlers': ['console'],
                    'level': 'DEBUG' if debug else 'INFO',
                },
            },
        },
    )


def start_worker_process(unsecure, runner_type):
    worker = Worker(secure=not unsecure, runner_type=runner_type)
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(
        signal.SIGINT,
        worker.handle_signal,
    )
    loop.add_signal_handler(
        signal.SIGTERM,
        worker.handle_signal,
    )
    loop.run_until_complete(worker.start())


def start(data):  # noqa: CCR001
    uvloop.install()
    logger.info('Starting Connect EaaS runtime....')
    if data.unsecure:
        logger.warning('Websocket connections will be established using unsecure protocol (ws).')

    if not data.split:
        start_worker_process(data.unsecure, None)
    else:
        workers = {}
        exited_workers = {}
        stop_event = threading.Event()

        for runner_type in ('interactive', 'background'):
            p = Process(daemon=True, target=start_worker_process, args=(data.unsecure, runner_type))
            workers[runner_type] = p
            exited_workers[runner_type] = False
            p.start()
            logger.info(f'{runner_type} tasks worker pid: {p.pid}')

        def _terminate(*args, **kwargs):  # pragma: no cover
            for p in workers.values():
                p.join()
            stop_event.set()

        signal.signal(signal.SIGINT, _terminate)
        signal.signal(signal.SIGTERM, _terminate)

        while not (stop_event.is_set() or all(exited_workers.values())):
            for runner_type, p in workers.items():
                if not p.is_alive():
                    if p.exitcode != 0:
                        notify_process_restarted(runner_type)
                        logger.info(f'Process of type {runner_type} is dead, restart it')
                        p = Process(
                            daemon=True,
                            target=start_worker_process,
                            args=(data.unsecure, runner_type),
                        )
                        workers[runner_type] = p
                        p.start()
                    else:
                        logger.info(f'worker {runner_type} exited')
                        exited_workers[runner_type] = True

            time.sleep(PROCESS_CHECK_INTERVAL_SECS)

        logger.info('Connect EaaS runtime terminated.')


def main():
    parser = argparse.ArgumentParser(prog='cextrun')
    parser.add_argument('-u', '--unsecure', action='store_true')
    parser.add_argument('-s', '--split', action='store_true', default=False)
    parser.add_argument('-d', '--debug', action='store_true', default=False)
    data = parser.parse_args()
    configure_logger(data.debug)
    start(data)


if __name__ == '__main__':  # pragma: no cover
    main()



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import inspect
import logging
import os
import subprocess
from datetime import datetime
from uuid import uuid4

from connect.client import ClientError, ConnectClient

from pkg_resources import (
    DistributionNotFound,
    get_distribution,
    iter_entry_points,
)

from connect.eaas.constants import (
    BACKGROUND_TASK_MAX_EXECUTION_TIME,
    INTERACTIVE_TASK_MAX_EXECUTION_TIME,
    ORDINAL_SUFFIX,
    SCHEDULED_TASK_MAX_EXECUTION_TIME,
    TASK_TYPE_EXT_METHOD_MAP,
)
from connect.eaas.exceptions import EaaSError


logger = logging.getLogger('connect.eaas')


def get_container_id():
    result = subprocess.run(
        ['cat', '/proc/1/cpuset'],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
    except subprocess.CalledProcessError:
        return str(uuid4())

    _, container_id = result.stdout.decode()[:-1].rsplit('/', 1)
    if len(container_id) == 64:
        return container_id

    result = subprocess.run(
        ['grep', 'overlay', '/proc/self/mountinfo'],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
        mount = result.stdout.decode()
        start_idx = mount.index('upperdir=') + len('upperdir=')
        end_idx = mount.index(',', start_idx)
        dir_path = mount[start_idx:end_idx]
        _, container_id, _ = dir_path.rsplit('/', 2)
        if len(container_id) != 64:
            return str(uuid4())
        return container_id
    except (subprocess.CalledProcessError, ValueError):
        return str(uuid4())


def get_environment():
    return {
        'api_key': os.getenv('API_KEY'),
        'environment_id': os.getenv('ENVIRONMENT_ID'),
        'instance_id': os.getenv('INSTANCE_ID', get_container_id()),
        'ws_address': os.getenv('SERVER_ADDRESS', 'api.cnct.info'),
        'api_address': os.getenv('API_ADDRESS', os.getenv('SERVER_ADDRESS', 'api.cnct.info')),
        'background_task_max_execution_time': int(os.getenv(
            'BACKGROUND_TASK_MAX_EXECUTION_TIME', BACKGROUND_TASK_MAX_EXECUTION_TIME,
        )),
        'interactive_task_max_execution_time': int(os.getenv(
            'INTERACTIVE_TASK_MAX_EXECUTION_TIME', INTERACTIVE_TASK_MAX_EXECUTION_TIME,
        )),
        'scheduled_task_max_execution_time': int(os.getenv(
            'SCHEDULED_TASK_MAX_EXECUTION_TIME', SCHEDULED_TASK_MAX_EXECUTION_TIME,
        )),
    }


def get_extension_class():
    ext_class = next(iter_entry_points('connect.eaas.ext', 'extension'), None)
    return ext_class.load() if ext_class else None


def get_extension_type(cls):
    descriptor = cls.get_descriptor()
    guess_async = [
        inspect.iscoroutinefunction(getattr(cls, TASK_TYPE_EXT_METHOD_MAP[name]))
        for name in descriptor['capabilities'].keys()
    ] + [
        inspect.iscoroutinefunction(getattr(cls, schedulable['method']))
        for schedulable in descriptor.get('schedulables', [])
    ]

    if all(guess_async):
        return 'async'
    if not any(guess_async):
        return 'sync'

    raise EaaSError('An Extension class can only have sync or async methods not a mix of both.')


def get_version():
    try:
        return get_distribution('connect-extension-runner').version
    except DistributionNotFound:
        return '0.0.0'


def to_ordinal(val):
    if val > 14:
        return f"{val}{ORDINAL_SUFFIX.get(int(str(val)[-1]), 'th')}"
    return f"{val}{ORDINAL_SUFFIX.get(val, 'th')}"


def get_client():
    env = get_environment()
    api_key = env['api_key']
    api_address = env['api_address']
    return ConnectClient(
        api_key,
        endpoint=f'https://{api_address}/public/v1',
        use_specs=False,
        max_retries=3,
    )


def get_current_environment():
    env = get_environment()
    environment_id = env['environment_id']
    if environment_id.startswith('ENV-'):
        service_id = f'SRVC-{environment_id[4:-3]}'
        client = get_client()
        try:
            return (
                client('devops').services[service_id].environments[environment_id].get()
            )
        except ClientError as ce:
            logger.warning(f'Cannot retrieve environment information: {ce}')


def notify_process_restarted(process_type):
    env = get_environment()
    current_environment = get_current_environment()
    if current_environment and current_environment['runtime'] == 'cloud':
        instance_id = env['instance_id']
        environment_id = current_environment['id']
        service_id = f'SRVC-{environment_id[4:-3]}'
        client = get_client()
        try:
            client('devops').services[service_id].environments[environment_id].update(
                {
                    'error_output': (
                        f'Process {process_type} worker of instance '
                        f'{instance_id} has been '
                        f'restarted at {datetime.now().isoformat()}'
                    ),
                },
            )
        except ClientError as ce:
            logger.warning(f'Cannot notify {process_type} process restart: {ce}')



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import asyncio
import dataclasses
import logging
import time
import traceback

from connect.eaas.constants import (
    ASSET_REQUEST_TASK_TYPES,
    LISTING_REQUEST_TASK_TYPES,
    TASK_TYPE_EXT_METHOD_MAP,
    TIER_CONFIG_REQUEST_TASK_TYPES,
)
from connect.eaas.dataclasses import (
    ResultType,
    TaskPayload,
    TaskType,
)
from connect.client import R
from connect.eaas.extension import ProcessingResponse
from connect.eaas.managers.base import TasksManagerBase


logger = logging.getLogger(__name__)


class BackgroundTasksManager(TasksManagerBase):

    def get_method(self, task_data, extension, argument):
        method_name = TASK_TYPE_EXT_METHOD_MAP[task_data.task_type]
        return getattr(extension, method_name, None)

    async def get_argument(self, task_data):
        """
        Get the request object through the Connect public API
        related to the task that need processing.
        """
        if task_data.task_type in ASSET_REQUEST_TASK_TYPES:
            return await self._get_asset_request(task_data)
        if task_data.task_type in TIER_CONFIG_REQUEST_TASK_TYPES:
            return await self._get_tier_config_request(task_data)
        if task_data.task_type in LISTING_REQUEST_TASK_TYPES:
            return await self._get_listing_request(task_data)
        if task_data.task_type == TaskType.TIER_ACCOUNT_UPDATE_REQUEST_PROCESSING:
            return await self._get_tier_account_request(task_data)
        if task_data.task_type == TaskType.USAGE_FILE_REQUEST_PROCESSING:
            return await self._get_usage_file(task_data)
        if task_data.task_type == TaskType.PART_USAGE_FILE_REQUEST_PROCESSING:
            return await self._get_usage_file_chunk(task_data)

    async def build_response(self, task_data, future):
        """
        Wait for a background task to be completed and than uild the task result message.
        """
        result_message = TaskPayload(**dataclasses.asdict(task_data))
        result = None
        try:
            begin_ts = time.monotonic()
            result = await asyncio.wait_for(
                future,
                timeout=self.config.get_timeout('background'),
            )
            result_message.result = result.status
            result_message.runtime = time.monotonic() - begin_ts
            logger.info(
                f'background task {task_data.task_id} result: {result.status}, tooks:'
                f' {result_message.runtime}',
            )
            if result.status in (ResultType.SKIP, ResultType.FAIL):
                result_message.output = result.output

            if result.status == ResultType.RESCHEDULE:
                result_message.countdown = result.countdown
        except Exception as e:
            self.log_exception(task_data, e)
            result_message.result = ResultType.RETRY
            result_message.output = traceback.format_exc()[:4000]

        return result_message

    def send_skip_response(self, data, output):
        future = asyncio.Future()
        future.set_result(ProcessingResponse.skip(output))
        asyncio.create_task(self.enqueue_result(data, future))

    async def _status_has_changed(self, task_data, namespace, collection, status_field):
        supported_statuses = self.handler.capabilities[task_data.task_type]
        client = self.client
        if namespace:
            client = client.ns(namespace)
        client = client.collection(collection)
        filter_expr = {
            'id': task_data.object_id,
            f'{status_field}__in': supported_statuses,
        }
        if await client.filter(**filter_expr).count() == 0:
            logger.info(
                f'Send skip response for {task_data.task_id} since '
                'the current request status is not supported.',
            )
            self.send_skip_response(
                task_data,
                'The request status does not match the '
                f'supported statuses: {",".join(supported_statuses)}.',
            )
            return True

        return False

    async def _get_asset_request(self, task_data):
        if await self._status_has_changed(task_data, None, 'requests', 'status'):
            return

        return await self.client.requests[task_data.object_id].get()

    async def _get_tier_config_request(self, task_data):
        if await self._status_has_changed(task_data, 'tier', 'config-requests', 'status'):
            return

        return await self.client('tier').config_requests[task_data.object_id].get()

    async def _get_listing_request(self, task_data):

        if await self._status_has_changed(task_data, None, 'listing-requests', 'state'):
            return

        listing_request = await self.client.listing_requests[task_data.object_id].get()

        if self.config.hub_id is None:
            return listing_request

        marketplace_id = listing_request['listing']['contract']['marketplace']['id']
        marketplace = await self.client.marketplaces[marketplace_id].get()
        hubs = [
            marketplace_hub['hub']['id']
            for marketplace_hub in marketplace['hubs']
        ]
        if self.config.hub_id not in hubs:
            self.send_skip_response(
                task_data,
                f'The marketplace {marketplace_id} does not belong '
                f'to hub {self.config.hub_id}.',
            )
            return

        return listing_request

    async def _get_tier_account_request(self, task_data):
        if await self._status_has_changed(task_data, 'tier', 'account-requests', 'status'):
            return

        tar = await self.client('tier').account_requests[task_data.object_id].get()
        tier_account_id = tar['account']['id']
        product_id = tar['product']['id']
        connection_type = (
            'preview' if self.config.environment_type == 'development'
            else self.config.environment_type
        )
        tiers_filter = (
            R().tiers.tier2.id.eq(tier_account_id)
            | R().tiers.tier1.id.eq(tier_account_id)
            | R().tiers.customer.id.eq(tier_account_id)
        )
        query = (
            R().product.id.eq(product_id)
            & R().connection.type.eq(connection_type)
            & tiers_filter
        )
        if await self.client.assets.filter(query).count() == 0:
            self.send_skip_response(
                task_data,
                'The Tier Account related to this request does not '
                f'have assets with a {connection_type} connection.',
            )
            return

        return tar

    async def _get_usage_file(self, task_data):
        if await self._status_has_changed(task_data, 'usage', 'files', 'status'):
            return

        return await self.client('usage').files[task_data.object_id].get()

    async def _get_usage_file_chunk(self, task_data):
        if await self._status_has_changed(task_data, 'usage', 'chunks', 'status'):
            return

        return await self.client('usage').chunks[task_data.object_id].get()



import logging

from connect.client import AsyncConnectClient, ConnectClient
from connect.eaas.helpers import (
    get_extension_class,
    get_extension_type,
)
from connect.eaas.config import ConfigHelper
from connect.eaas.logging import ExtensionLogHandler, RequestLogger


class ExtensionHandler:
    """
    Handle the lifecycle of an extension.
    """
    def __init__(self, config: ConfigHelper):
        self.config = config
        self.extension_class = get_extension_class()
        self.extension_type = get_extension_type(self.extension_class)
        self.descriptor = self.extension_class.get_descriptor()
        self.logging_handler = None

    @property
    def capabilities(self):
        return self.descriptor['capabilities']

    @property
    def variables(self):
        return self.descriptor.get('variables')

    @property
    def schedulables(self):
        return self.descriptor.get('schedulables')

    @property
    def readme(self):
        return self.descriptor['readme_url']

    @property
    def changelog(self):
        return self.descriptor['changelog_url']

    def new_extension(self, task_id):
        return self.extension_class(
            self._create_client(task_id),
            logging.LoggerAdapter(
                self._create_logger(),
                {'task_id': task_id},
            ),
            self.config.variables,
        )

    def _create_logger(self):
        """
        Returns a logger instance configured with the LogZ.io handler.
        This logger will be used by the extension to send logging records
        to the Logz.io service.
        """
        logger = logging.getLogger('eaas.extension')
        if self.logging_handler is None and self.config.logging_api_key is not None:
            self.logging_handler = ExtensionLogHandler(
                self.config.logging_api_key,
                default_extra_fields=self.config.metadata,
            )
            logger.addHandler(self.logging_handler)
        return logger

    def _create_client(self, task_id):
        """
        Get an instance of the Connect Openapi Client. If the extension is asyncrhonous
        it returns an instance of the AsyncConnectClient otherwise the ConnectClient.
        """
        Client = ConnectClient if self.extension_type == 'sync' else AsyncConnectClient
        return Client(
            self.config.api_key,
            endpoint=self.config.get_api_url(),
            use_specs=False,
            logger=RequestLogger(
                logging.LoggerAdapter(
                    self._create_logger(),
                    {'task_id': task_id},
                ),
            ),
            default_headers=self.config.get_user_agent(),
        )



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import json

import pkg_resources

from connect.eaas.dataclasses import ResultType


class _Response:
    def __init__(self, status):
        self.status = status


class ProcessingResponse(_Response):

    def __init__(self, status, countdown=30, output=None):
        super().__init__(status)
        self.countdown = 30 if countdown < 30 else countdown
        self.output = output

    @classmethod
    def done(cls):
        return cls(ResultType.SUCCESS)

    @classmethod
    def skip(cls, output=None):
        return cls(ResultType.SKIP, output=output)

    @classmethod
    def reschedule(cls, countdown=30):
        return cls(ResultType.RESCHEDULE, countdown=countdown)

    @classmethod
    def slow_process_reschedule(cls, countdown=300):
        return cls(
            ResultType.RESCHEDULE,
            countdown=300 if countdown < 300 else countdown,
        )

    @classmethod
    def fail(cls, output=None):
        return cls(ResultType.FAIL, output=output)


class ValidationResponse(_Response):
    def __init__(self, status, data, output=None):
        super().__init__(status)
        self.data = data
        self.output = output

    @classmethod
    def done(cls, data):
        return cls(ResultType.SUCCESS, data)

    @classmethod
    def fail(cls, data=None, output=None):
        return cls(ResultType.FAIL, data=data, output=output)


class _InteractiveTaskResponse(_Response):
    def __init__(self, status, http_status, headers, body, output):
        super().__init__(status)
        self.http_status = http_status
        self.headers = headers
        self.body = body
        self.output = output

    @property
    def data(self):
        return {
            'http_status': self.http_status,
            'headers': self.headers,
            'body': self.body,
        }

    @classmethod
    def done(cls, http_status=200, headers=None, body=None):
        return cls(ResultType.SUCCESS, http_status, headers, body, None)

    @classmethod
    def fail(cls, http_status=400, headers=None, body=None, output=None):
        return cls(ResultType.FAIL, http_status, headers, body, output)


class CustomEventResponse(_InteractiveTaskResponse):
    pass


class ProductActionResponse(_InteractiveTaskResponse):
    pass


class ScheduledExecutionResponse(_Response):

    def __init__(self, status, output=None):
        super().__init__(status)
        self.output = output

    @classmethod
    def done(cls):
        return cls(ResultType.SUCCESS)

    @classmethod
    def fail(cls, output=None):
        return cls(ResultType.FAIL, output=output)


class Extension:
    def __init__(self, client, logger, config):
        self.client = client
        self.logger = logger
        self.config = config

    @classmethod
    def get_descriptor(cls):  # pragma: no cover
        return json.load(
            pkg_resources.resource_stream(
                cls.__module__,
                'extension.json',
            ),
        )

    def process_asset_purchase_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_asset_change_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_asset_suspend_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_asset_resume_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_asset_cancel_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_asset_adjustment_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def validate_asset_purchase_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def validate_asset_change_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_tier_config_setup_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_tier_config_change_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_tier_config_adjustment_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def validate_tier_config_setup_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def validate_tier_config_change_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def execute_product_action(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_product_custom_event(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_new_listing_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_remove_listing_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_tier_account_update_request(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_usage_file(self, request):  # pragma: no cover
        raise NotImplementedError()

    def process_usage_chunk_file(self, request):  # pragma: no cover
        raise NotImplementedError()
